--[[
Name: LibQuestInfo
Revision: $Revision: 1 $
Author(s): Angelike
Website:
Documentation: www.wowace.com
SVN:
Description: Information about quests
License: LGPL v2.1
]]

do -- check for missing libraries
	local MissingLibraries
	if not LibStub then
		print("Libstub not found, LibQuestInfo not loaded.")
		return;
	end
	local REQUIRED_LIBS = {
		"CallbackHandler-1.0",
	}
	local missing

	for idx, lib in ipairs(REQUIRED_LIBS) do
		if not LibStub:GetLibrary(lib, true) then
			missing = true
			print(format("The Library %s could not be found LibQuestInfo not loaded.", lib))
		end
	end
	if missing then
		return
	end
end

local MAJOR_VERSION, MINOR_VERSION = "LibQuestInfo", tonumber(("$Revision: 1 $"):match("%d+")) or 0;
local lib, oldMinor = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION);
if not lib then return end;

do -- table functions extended
	local list = setmetatable({}, {__mode='k'})
	table.new = function(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	table.del = function(t)
		if type(t) ~= "table" then
			return nil
		end
		setmetatable(t, nil)
		table.wipe(t)
		list[t] = true
		return nil
	end
	table.deepDel = function(t)
		if type(t) ~= "table" then
			return nil
		end
		for k,v in pairs(t) do
			t[k] = table.deepDel(v)
		end
		return table.del(t)
	end
	table.doAll = function(func, ...)
		for i=1, select('#', ...) do
			table[func](select(i, ...))
		end
	end
	table.copy = function (t)
	   local mt =   getmetatable(t);
	   local tn = {};
	   for k, v in pairs(t) do
		  table.insert(tn, k, v);
	   end
	   if mt then
		  setmetatable(tn, mt);
	   end
	   return tn;
	end
end

lib.events = lib.events or LibStub("CallbackHandler-1.0"):New(lib)

do --create sub  tables for update stuff
	-- debug
	_G['LibQuestInfo'] = lib;
	lib.updateQuestTime = 0.1;
	lib.updateFrameTime = 0;
	lib.updateToolTipTime = 0;
	lib.updateGroupTime = 0.

	lib.db = table.new();
	lib.db.party = table.new();
	lib.db.Cache = {
		["QuestCache"] = table.new(),
		["CategoryCache"] = table.new(),
		["QuestCategoryCache"] = table.new(), -- cache for questid's in zones, should only used by questeditor
	}
	lib.db.UpdateQueue = table.new(); 
	lib.db.SortCache = table.new(); 

	--[[Possible Zone Datakeys:
	A	:	AchievementID
	AC	:	Achievement:CriteriaNr
	CD	:	CriteriaID
	C	:	CathegorieID
	Z	:	ZoneID
	--]]	
	lib.db.CategoryData = {
		--[[ primary profession Achievement
		[] = 5551, -- Alchemy
		[] = 5552, -- Blacksmithing
		[] = 5553, -- Encharnting
		[] = 5554, -- Engeneering
		[] = 5555, -- Herbalism
		[] = 5556, -- Inscription
		[] = 5557, -- Jewelcrafting
		[] = 5558, -- Leatherworking
		[] = 5559, -- Mining
		[] = 5560, -- Skinning
		[] = 5561, -- Tailoring

		-- Secondary Profession
		[] = 5562, -- First Aid
		[] = 5563, -- Cooking
		[] = 5564, -- Fishing
		[] = 14593, -- Archaeology

		-- Races
		[] = "A:246:1", -- bloodelf
		[] = "A:246:2", -- orc
		[] = "A:246:3", -- tauren
		[] = "A:246:4", -- undead
		[] = "A:246:5", -- troll
		[] = , -- goblin

		[] = "A:1005:1", -- draenei
		[] = "A:1005:2", -- human
		[] = "A:1005:3", -- gnome
		[] = "A:1005:4", -- dwarf
		[] = "A:1005:5", -- night elf
		[] = , -- worgen
		
		[] = , -- pandaren
		--

		-- Classes
		[] = "AC:4989:1", -- Death Knight
		[] = "AC:4989:2", -- Druid
		[] = "AC:4989:3", -- Hunter
		[] = "AC:4989:4", -- Mage
		[] = "AC:4989:5", -- Paladin
		[] = "AC:4989:6", -- Priest
		[] = "AC:4989:7", -- Rouge
		[] = "AC:4989:8", -- Shaman
		[] = "AC:4989:9", -- Warlock
		[] = "AC:4989:10", -- Warrior

		-- Battlegrounds
		[] = "C:153", -- Battlegrounds
		[] = "C:14804", -- Warsong
		[] = "C:14802", -- Arathi Basin
		[] = "C:14801", -- Alterac Valley
		[] = "C:14803", -- Eye of the Storm
		[] = "C:14881", -- Strand of the Ancients
		[] = "C:14901", -- Wintergrasp
		[] = "C:15003", -- Isle of Conquest
		[] = "C:15073", -- Battle for Gilneas
		[] = "C:15074", -- Twin Peaks
		[] = "C:15075", -- Tol Barad


		-- 101 - 200 Zones Eastern Kingdoms
		[] = "Z:14", -- Azeroth
		[] = "Z:23", -- EasternPlaguelands
		[] = "Z:22", -- WesternPlaguelands
		[] = "Z:20", -- Tirisfal
		[] = "Z:21", -- Silverpine
		[] = "Z:24", -- Hilsbrad
		[] = "Z:15", -- Alterac (ingame anymore?)
		[] = "Z:26", -- Hinterlands
		[] = "Z:16", -- Arathi
		[] = "Z:40", -- Wetlands
		[] = "Z:35", -- LochModan
		[] = "Z:27", -- DunMorogh
		[] = "Z:28", -- SearingGorge
		[] = "Z:17", -- Badlands
		[] = "Z:29", -- BurningSteppes
		[] = "Z:36", -- Redridge
		[] = "Z:30", -- Elwynn
		[] = "Z:39", -- Westfall
		[] = "Z:34", -- Duskwood
		[] = "Z:32", -- DeadwindPass
		[] = "Z:38", -- SwampOfSorrows
		[] = "Z:19", -- BlastedLands
		[] = "Z:37", -- Stranglethorn
		-- Capital Citys
		[] = "Z:301", -- Stormwind
		[] = "Z:341", -- Ironforge
		[] = "Z:382", -- Undercity
	 
		-- 201 - 300 Zones Kalimdor
		[] = "Z:13", -- Kalimdor
		[] = "Z:41", -- Teldrassil
		[] = "Z:42", -- Darkshore
		[] = "Z:241", -- Moonglade
		[] = "Z:281", -- Winterspring
		[] = "Z:182", -- Felwood
		[] = "Z:43", -- Ashenvale
		[] = "Z:181", -- Aszhara
		[] = "Z:4", -- Durotar
		[] = "Z:11", -- Barrens (ingame anymore in CC?)
		[] = "Z:81", -- StonetalonMountains
		[] = "Z:101", -- Desolace
		[] = "Z:9", -- Mulgore
		[] = "Z:141", -- Dustwallow
		[] = "Z:61", -- ThousandNeedles
		[] = "Z:121", -- Feralas
		[] = "Z:261", -- Silithus
		[] = "Z:201", -- UngoroCrater
		[] = "Z:161", -- Tanaris
		-- Capital Citys
		[] = "Z:321", -- Ogrimmar
		[] = "Z:362", -- ThunderBluff
		[] = "Z:381", -- Darnassis
	 
		-- 301 - 350 Zones BC
		[] = "Z:466", -- Expansion01
		[] = "Z:499", -- Sunstrider Island ??
		[] = "Z:462", -- EversongWoods
		[] = "Z:463", -- Ghostlands
		[] = "Z:480", -- SilvermoonCity
		[] = "Z:464", -- AzuremystIsle
		[] = "Z:476", -- BloodmystIsle
		[] = "Z:471", -- TheExodar
		[] = "Z:479", -- Netherstorm
		[] = "Z:475", -- BladesEdgeMountains
		[] = "Z:465", -- Hellfire
		[] = "Z:467", -- Zangarmarsh
		[] = "Z:477", -- Nagrand
		[] = "Z:478", -- TerokkarForest
		[] = "Z:473", -- ShadowmoonValley
		[] = "Z:481", -- ShattrathCity

		-- 351 - 400 Zones LK
		[] = "Z:485", -- Northrend
		[] = "Z:502", -- ScarletEnclave
		[] = "Z:491", -- HowlingFjord
		[] = "Z:490", -- GrizzlyHills
		[] = "Z:488", -- Dragonblight
		[] = "Z:501", -- LakeWintergrasp
		[] = "Z:486", -- BoreanTundra
		[] = "Z:493", -- SholazarBasin
		[] = "Z:492", -- IcecrownGlacier
		[] = "Z:510", -- CrystalsongForest
		[] = "Z:496", -- ZulDrak
		[] = "Z:495", -- TheStormPeaks
		[] = "Z:541", -- HrothgarsLanding
		[] = "Z:504", -- Dalaran

		-- 401 - 420 Zones CC
		[] = "(PH) Gilneas", -- Gilneas
		[] = "(PH) Tol Barad", -- Tol Barad
		[] = "(PH) Twilight Highlands", -- Twilight Highlands
		[] = "(PH) Vashj'ir", -- Vashj'ir
		[] = "(PH) Kelp'thar Forest", -- Kelp'thar Forest
		[] = "(PH) Shimmering Expanse", -- Shimmering Expanse
		[] = "(PH) Abyssal Depths", -- Abyssal Depths
		[] = "(PH) Gilneas City", -- Gilneas City
		[] = "(PH) Ruins of Gilneas", -- Ruins of Gilneas
		[] = "(PH) Ruins of Gilneas City", -- Ruins of Gilneas City
		[] = "(PH) Mount Hyjal", -- Mount Hyjal
		[] = "Z:720", -- Uldum
		[] = "(PH) Kezan", -- Kezan
		[] = "(PH) The Lost Islands", -- The Lost Islands
		[] = "(PH) Deepholm", -- Deepholm

		-- 501 - 510 Raids Vanilla (7?)
		[] = "A:686", -- MoltenCore
		[] = "A:685", -- BlackwingLair
		[] = "(PH) Naxx", -- Naxxramas
		[] = "A:688", -- Zul Gurub
		[] = "A:687", -- AhnQiraj
		[] = "A:689", -- RuinsofAhnQiraj
		[] = "Z:718", -- OnyxiasLair
		[] = "A:1307", -- Upper Blackrock Spire

		-- 511 - 520 Raids BC (9)
		[] = "A:690", -- Karazhan
		[] = "A:691", -- Zul'Aman
		[] = "A:692", -- Gruul's lair
		[] = "A:693", -- Megtheridon's Lair
		[] = "A:694", -- Serpentshrine Cavern
		[] = "A:695", -- The Battle for Mount Hyjal
		[] = "A:696", -- Tempest Keep
		[] = "A:697", -- The Black Temple
		[] = "A:698", -- Sunwell Plateau

		-- 521 - 530 Raids LK (7)
		[] = "Z:535", -- Naxxramas
		[] = "Z:532", -- VaultofArchavon
		[] = "Z:527", -- TheEyeofEternity
		[] = "Z:531", -- TheObsidianSanctum
		[] = "Z:529", -- Ulduar
		[] = "Z:542", -- TheArgentColiseum
		[] = "Z:609", -- RubySanctum
		[] = "(PH) Icecrown Raid", -- Icecrown Raid

		-- 531 - 540 Raids CC
		[] = "Z:752", -- BardinHold (Tol Barad Raid)
		[] = "Z:754", -- BlackwingDescent
		[] = "Z:758", -- BastionOfTwilight
		[] = "Z:773", -- ThroneOfTheFourWinds

		-- 551 - 570 Dungeons Vanilla WoW
		[] = "A:628", -- Deadmines
		[] = "A:629", -- Ragefire Chasm
		[] = "A:630", -- Wailing Caverns
		[] = "A:631", -- Shadowfang Keep
		[] = "A:632", -- Blackthom Deeps
		[] = "A:633", -- Stormwind Stockade
		[] = "A:634", -- Gnomerigan
		[] = "A:635", -- Razorfen Kraul
		[] = "A:636", -- Razorfen Downs
		[] = "A:637", -- Scarlet Monastery
		[] = "A:638", -- Uldaman
		[] = "A:639", -- Zul'Farrak
		[] = "A:640", -- Maraudon
		[] = "A:641", -- Sunken Temple
		[] = "A:642", -- Blackrock Deeps
		[] = "A:643", -- Lower Blackrock Spire
		[] = "A:644", -- Dire Maul --ToDo
		[] = "A:645", -- Scholomance
		[] = "A:646", -- Stratholme

		-- 601 - 620 Dungeons BC
		[] = "A:647", -- Hellfire Ramparts
		[] = "A:648", -- The Blood Furnace
		[] = "A:649", -- The Slave Pens
		[] = "A:650", -- Underbog
		[] = "A:651", -- Mana-Tombs
		[] = "A:652", -- The Escape From Durnholde
		[] = "A:653", -- Sethekk Halls
		[] = "A:654", -- Shadow Labyrinth
		[] = "A:655", -- Opening of the Dark Portal
		[] = "A:656", -- The Steamvault
		[] = "A:657", -- The Shattered Halls
		[] = "A:658", -- The Mechanar
		[] = "A:659", -- The Botanica
		[] = "A:660", -- The Arcatraz
		[] = "A:661", -- Magister's Terrace
		[] = "A:666", -- Auchenai Crypts

		-- Dungeon LK
		[] = "A:477", -- Utgarde Keep
		[] = "A:478", -- The Nexus
		[] = "A:479", -- The Culling of Stratholme
		[] = "A:480", -- Azjol-Nerub
		[] = "A:481", -- Ahn'kahet: The Old Kingdom
		[] = "A:482", -- Drak'Tharon Keep
		[] = "A:483", -- The Violet Hold
		[] = "A:484", -- Gundrak
		[] = "A:485", -- Halls of Stone
		[] = "A:486", -- Halls of Lightning
		[] = "A:487", -- The Oculus
		[] = "A:488", -- Utgarde Pinnacle
		[] = "A:4296", -- Trial of the Champion
		[] = "A:4516", -- Forge of Souls
		[] = "A:4517", -- Pit of Saron
		[] = "A:4518", -- Halls of Reflection
		-- todo: add patch dungeons

		-- Dungeons CC
		[] = "A:4833", -- Blackrock Caverns
		[] = "A:4840", -- Grim Batol
		[] = "A:4841", -- Halls of Origination
		[] = "A:4848", -- Lost City of the Tol'vir
		[] = "A:4846", -- The Stonecore
		[] = "A:4847", -- The Vortex Pinnacle
		[] = "A:4839", -- Throne of the Tides
		
		-- Pandaria
		-- Zones
		--]]

	}
	lib.db.ExpData = {
		[1] = "10|20|40|60|80|100|120|160",
		[2] = "15|45|85|130|170|215|255|340",
		[3] = "25|65|125|190|250|315|380|500",
		[4] = "35|90|180|270|355|445|540|710",
		[5] = "45|110|225|335|450|560|670|900",
		[6] = "55|135|270|405|540|680|810|1080",
		[7] = "65|160|315|475|630|790|950|1260",
		[8] = "70|175|350|530|700|880|1050|1400",
		[9] = "80|195|390|590|780|980|1150|1560",
		[10] = "85|210|420|630|840|1050|1250|1680",
		[11] = "90|220|440|660|880|1100|1300|1760",
		[12] = "90|225|455|680|910|1150|1350|1820",
		[13] = "90|230|455|680|910|1150|1350|1820",
		[14] = "100|245|490|740|980|1250|1450|1960",
		[15] = "105|270|540|800|1050|1350|1600|2100",
		[16] = "115|290|580|880|1150|1450|1750|2300",
		[17] = "125|315|630|950|1250|1600|1900|2500",
		[18] = "135|340|680|1000|1350|1700|2050|2700",
		[19] = "145|365|730|1100|1450|1800|2200|2900",
		[20] = "155|390|780|1150|1550|1950|2350|3100",
		[21] = "165|415|830|1250|1650|2050|2500|3300",
		[22] = "175|435|870|1300|1750|2200|2600|3500",
		[23] = "185|460|920|1400|1850|2300|2750|3700",
		[24] = "195|485|970|1450|1950|2400|2900|3900",
		[25] = "200|510|1000|1500|2000|2550|3050|4000",
		[26] = "210|530|1050|1600|2100|2650|3150|4200",
		[27] = "220|550|1100|1650|2200|2750|3300|4400",
		[28] = "230|570|1150|1700|2300|2850|3400|4600",
		[29] = "235|590|1200|1750|2350|2950|3550|4700",
		[30] = "245|610|1200|1850|2450|3050|3650|4900",
		[31] = "250|630|1250|1900|2540|3150|3800|5080",
		[32] = "270|670|1350|2000|2710|3350|4050|5420",
		[33] = "290|720|1450|2150|2900|3600|4350|5800",
		[34] = "310|770|1550|2300|3100|3850|4650|6200",
		[35] = "330|820|1650|2450|3300|4100|4950|6600",
		[36] = "350|870|1750|2600|3500|4350|5250|7000",
		[37] = "370|920|1850|2750|3710|4600|5550|7420",
		[38] = "390|980|1950|2900|3920|4900|5850|7840",
		[39] = "410|1030|2050|3100|4140|5150|6200|8280",
		[40] = "435|1090|2150|3250|4370|5450|6550|8740",
		[41] = "455|1140|2250|3400|4590|5700|6850|9180",
		[42] = "480|1200|2400|3600|4820|6000|7200|9640",
		[43] = "505|1260|2500|3750|5050|6300|7550|10100",
		[44] = "525|1320|2600|3950|5290|6600|7900|10580",
		[45] = "550|1380|2750|4150|5540|6900|8300|11080",
		[46] = "575|1440|2850|4300|5790|7200|8650|11580",
		[47] = "600|1510|3000|4500|6040|7550|9050|12080",
		[48] = "625|1570|3100|4700|6290|7850|9400|12580",
		[49] = "655|1630|3250|4900|6550|8150|9800|13100",
		[50] = "680|1700|3400|5100|6810|8500|10200|13620",
		[51] = "705|1760|3500|5300|7070|8800|10600|14140",
		[52] = "730|1830|3650|5500|7340|9150|11000|14680",
		[53] = "760|1900|3800|5700|7610|9500|11400|15220",
		[54] = "785|1970|3900|5900|7890|9850|11800|15780",
		[55] = "815|2040|4050|6100|8170|10200|12250|16340",
		[56] = "845|2110|4200|6300|8450|10550|12650|16900",
		[57] = "870|2180|4350|6500|8730|10900|13050|17460",
		[58] = "900|2250|4500|6750|9020|11250|13500|18040",
		[59] = "930|2320|4650|6950|9310|11600|13950|18620",
		[60] = "955|2380|4750|7150|9550|11900|14300|19100",
		[61] = "970|2400|4850|7300|9800|12200|14700|19600",
		[62] = "1000|2550|5000|7600|10050|12600|15050|20100",
		[63] = "1050|2600|5250|7800|10400|12950|15550|20800",
		[64] = "1080|2650|5400|8050|10750|13350|16000|21500",
		[65] = "1100|2700|5500|8250|11000|13750|16500|22000",
		[66] = "1130|2850|5650|8550|11300|14150|17000|22600",
		[67] = "1160|2900|5800|8750|11650|14600|17450|23300",
		[68] = "1200|3050|6000|9000|12000|14950|17950|24000",
		[69] = "1230|3100|6150|9250|12300|15400|18450|24600",
		[70] = "1250|3150|6250|9500|12650|15800|19000|25300",
		[71] = "2000|5050|10050|15100|20100|25150|30150|40200",
		[72] = "2050|5100|10150|15250|20300|25400|30450|40600",
		[73] = "2050|5150|10250|15400|20500|25650|30750|41000",
		[74] = "2100|5200|10400|15550|20750|25950|31150|41500",
		[75] = "2100|5250|10500|15700|20950|26200|31450|41900",
		[76] = "2100|5300|10600|15850|21150|26450|31750|42300",
		[77] = "2150|5350|10700|16050|21400|26750|32100|42800",
		[78] = "2150|5400|10800|16200|21600|27000|32400|43200",
		[79] = "2200|5450|10900|16350|21800|27250|32700|43600",
		[80] = "2200|5500|11050|16550|22050|27550|33100|44100",
		[81] = "2770|6950|13850|20800|27700|34650|41550|55400",
		[82] = "3490|8750|17450|26200|34900|43650|52350|69800",
		[83] = "4390|11000|21950|32950|43900|54900|65850|87800",
		[84] = "5520|13800|27600|41400|55200|69000|82800|110400",
		[85] = "6940|17350|34700|52050|69400|86750|104100|138800",
		[86] = "7190|18000|35950|53950|71900|89900|107850|143800",
		[87] = "7440|18600|37200|55800|74400|93000|111600|148800",
		[88] = "7700|19250|38500|57750|77000|96250|115500|154000",
		[89] = "17500|43750|87500|131250|175000|218750|262500|350000",
		[90] = "17635|44100|88200|132250|176350|220450|264550|352700",
		[91] = "17765|44400|88850|133250|177650|222050|266500|355300",
		[92] = "17900|44750|89500|134250|179000|223750|268500|358000",
		[93] = "18030|45100|90150|135250|180300|225400|270450|360600",
		[94] = "18160|45400|90800|136200|181600|227000|272400|363200",
		[95] = "18290|45750|91450|137200|182900|228650|274350|365800",
		[96] = "18420|46050|92100|138150|184200|230250|276300|368400",
		[97] = "18545|46350|92750|139100|185450|231800|278200|370900",
		[98] = "18675|46700|93400|140050|186750|233450|280150|373500",
		[99] = "18800|47000|94000|141000|188000|235000|282000|376000",
		[100] = "18930|47350|94650|142000|189300|236650|283950|378600",
	}
	lib.db.fractions = table.new();
end
-- #AUTODOC_NAMESPACE lib
local datakeys = {"O","L","M","H","Y","W","T","S","Z"}
--[[
	Datatypes:
	Q	=	QuestID + number
	O(#)	= # of ojectivetype
		-P	= Player
		-I 	= Item
		-M	= Monster
		-R	= Reputation
		-O	= Object
		-E	= Event
	L	=	Level + number
	l	=	Accept Level + number
	T	=	QuestTag
		P	=	PvP
		R	=	Raid
		D	=	Dungeon
		G	=	Group + number
		E	=	Elite
		H	=	Heroic
		Y	=	Daily
		W	=	Weekly
	Z	=	ZoneID
	S	=	Source + version (which is lib version number)
		S = Scan = version + 1
		P = Client Exchange = version + 0
		E = Editor = version - 1
	R(id)/(amount)	=	id of fraction / amount of gein . loss
--]]
-- Tag Type DB
local TagTypes = {
	P	=	PVP,
	R	=	RAID,
	D	=	LFG_TYPE_DUNGEON,
	G	=	GROUP,
	E	=	ELITE,
	H	=	ITEM_HEROIC,
}

-- Data Stuff done, init lib stuff

local _G, table, string, tonumber, tostring = _G, table, string, tonumber, tostring

local function print(args)
	if args ~= nil then
		_G["print"]("LQI: " .. tostring(args));
	end
end


if lib.frame then
	lib.frame:UnregisterAllEvents();
	lib.frame:SetScript("OnEvent", nil);
	lib.frame:SetScript("OnUpdate", nil);
else
	lib.frame = CreateFrame("Frame", MAJOR_VERSION .. "_Frame");
	lib.frame.doQuestLogUpdate = true;
end

if lib.callbacks then
	lib:UnregisterAll(lib) -- unregisters all callbacks that questinfo is doing on itself (comms stuff, mostly)
	lib:UnhookDialogs()
else
	lib.callbacks = LibStub("CallbackHandler-1.0"):New(lib, nil, nil, "UnregisterAll")
end

do -- Frame Registration
	--lib.frame:RegisterAllEvents();
	lib.frame:RegisterEvent("QUEST_ACCEPTED") -- unfinished
	--lib.frame:RegisterEvent("QUEST_LOG_UPDATE") -- unfinished
	--lib.frame:RegisterEvent("QUEST_COMPLETE") -- unfinished
	-- better the QUEST_LOG_UPDATE
	-- will only be fired if questlog changend and not when open/close header etc ...
	lib.frame:RegisterEvent("UNIT_QUEST_LOG_CHANGED") -- unfinished
	lib.frame:RegisterEvent("CHAT_MSG_ADDON") -- unfinished
	lib.frame:RegisterEvent("CHAT_MSG_SYSTEM") -- unfinished
	lib.frame:RegisterEvent("PARTY_MEMBERS_CHANGED") -- unfinished
	--lib.frame:RegisterEvent("PARTY_MEMBER_ENABLE") -- unfinished
	--lib.frame:RegisterEvent("PARTY_MEMBER_DISABLE") -- unfinished
	lib.frame:RegisterEvent("PARTY_CONVERTED_TO_RAID") -- unfinished
	lib.frame:RegisterEvent("RAID_ROSTER_UPDATE") -- unfinished
end

do -- Tooltip stuff
	if not tooltip then
		local tooltip = CreateFrame("GameTooltip", MAJOR_VERSION.."_Tooltip", lib.frame, "GameTooltipTemplate")
		tooltip:SetOwner(UIParent, "ANCHOR_NONE")
		lib.tooltip = tooltip
	end
	local function tooltip_line(link, line)
		tooltip:ClearLines()
		tooltip:SetHyperlink(link)

		if tooltip:NumLines() < line then return false end
		return _G[MAJOR_VERSION.."_TooltipTextLeft"..line]:GetText()
	end
	lib.db.UID_to_name = setmetatable(lib.db.UID_to_name or table.new(), {__index = function(self, key)
		local link = (type(key) == 'string') and key or ('quest:'..key)
		local uid = string.match(link, '%d+')
		local name = tooltip_line(link, 1)
		if name then
			self[uid] = name
			return name
		end
		return false
	end,})
end

--
-- Events
--

lib.frame:SetScript("OnEvent", function(this, event, ...)
	if (lib["internal_" .. event]) then
		lib["internal_" .. event](nil, this, ...);
		return
	end;
	
	if (string.find(event, "PARTY") or string.find(event, "RAID")) then
		-- toggle group Timer to make shure env vars arte updated correct
		this.updateGroupTime = this.updateGroupTime ~= nil and this.updateGroupTime >= 0 and this.updateGroupTime or 0.10;
		this.doGroupUpdate = true;
		return;
	end	
	
	print("Function not found: LibQuestInfo:internal_" .. event .. "()")
end)

do -- abandom stuff
	if not lib.hooked_abandomed then
		hooksecurefunc("AbandonQuest", function()
			-- set flag for Abandomed quest
			lib.loststatus = false;
		end)
		lib.hooked_abandomed = true;
	end
end

function lib:internal_UNIT_QUEST_LOG_CHANGED(frame, unitname, ...)
	if (unitname ~= nil and unitname ~= "player") then
		return;
	end
	-- schedule update timer
	frame.doQuestLogUpdate = true;
	frame.QuestLogUpdateTime = frame.QuestLogUpdateTime >= 0 and frame.QuestLogUpdateTime or 0.25;
end

function lib:internal_QUEST_ACCEPTED(frame, index, uid)
	lib:AddQuest(uid, GetUnitName("player"));
end

function lib:internal_QUEST_COMPLETE()
	-- Build String for last completition Quest
	lib.lastquest = string.format(ERR_QUEST_COMPLETE_S , GetTitleText());
end

function lib:internal_CHAT_MSG_SYSTEM(frame, chatMSG, ...)
	if (chatMSG == lib.lastquest) then
		-- last quest was quest from CompleteLog
		lib.loststatus = true;
		lib.lastquest = nil;
		return;
	end
end

function lib:internal_CHAT_MSG_ADDON(frame, prefix, msg, distribution, sender)
	if prefix ~= "LQI" or lib.disable_comms then 
		return
	end
	
	local commtype, remainder = msg:match("^(%a)(.*)")
	if (not commtype) or (not commhandlers[commtype]) then return end

	commhandlers[commtype](remainder, distribution, sender)
end
--
-- Events end
--

--
-- OnUpdate Stuff
--

-- this is just a throttler for QUEST_LOG_UPDATE
local timeSoFar = 0;
local failload = table.new();
lib.frame:SetScript("OnUpdate", function(this, elapsed)

	-- quest log update
	if this.doQuestLogUpdate then
		if not this.QuestLogUpdateTime then
			-- first run after login, schedule timer
			this.QuestLogUpdateTime = 0.25
		else
			this.QuestLogUpdateTime = this.QuestLogUpdateTime - elapsed;
		end
		if this.QuestLogUpdateTime < 0 then
			this.doQuestLogUpdate = nil;
			lib:internal_QUEST_LOG_UPDATE();
		end
	end
	
	-- Update grouplist
	if this.doGroupUpdate then
		if this.updateGroupTime >= 0 then
			this.updateGroupTime = this.updateGroupTime - elapsed;
		else
			this.doGroupUpdate = nil;
			lib:internal_DETECT_GROUP(this);
		end
	
	end
	
	
	-- quest stuff
	if lib.ProcessUpdatedQuestData then
		lib.updateQuestTime = lib.updateQuestTime - elapsed;
		if lib.updateQuestTime < 0 then
			-- tot load quest stuff
			lib.updateQuestTime = 0.1;
		end
	end
	
	-- zone stuff
	
	-- exp stuff
	
	timeSoFar = timeSoFar + elapsed
	local first, found
	if not (timeSoFar > 0.25) then
		return
	end

	do return end
	local currentid = lib.tooltip.currentid

	for i = 1, lib.tooltip:NumLines() do
		if not lib.tooltip[i] then
			lib.tooltip[i] = _G[lib.tooltip:GetName() .. "TextLeft" .. i]
		end
	end

	if currentid then
		if not questcache[currentid] then
			questcache[currentid] = table.new()
		end
		questcache[currentid].title = lib.tooltip[1] and lib.tooltip[1]:GetText()
		if not questcache[currentid].title then
			failload[currentid] = failload[currentid] and failload[currentid] + 1 or 1;
		end
		if (questcache[currentid].title) ~= nil then
			loadqueue[currentid] = nil
			lib.callbacks:Fire("QuestLoaded", currentid)
		end
	end

	timeSoFar = 0
	local size = 0
	for k in pairs(loadqueue) do
		if failload[currentid] and failload[currentid] > 10 then
			questcache[currentid].title = "Quest doesn't exist (Quest ID: " .. currentid ..")";
			loadqueue[k] = nil;
		else
			size = size + 1
		end
	end

	lib.tooltip:ClearLines();
	lib.tooltip:SetOwner(WorldFrame, "ANCHOR_NONE");

	if size > 0 then
		for qid in pairs(loadqueue) do
			if not lib.tooltip.currentid then
				lib.tooltip.currentid = qid
				lib.tooltip:SetHyperlink("quest:" .. qid)
				return
			end
			if not first then
				first = qid
			end
			if found then
				lib.tooltip.currentid = qid
				lib.tooltip:SetHyperlink("quest:" .. qid)
				return
			end

			if qid == lib.tooltip.currentid then
				found = true
			end
		end
		lib.tooltip.currentid = first
		lib.tooltip:SetHyperlink("quest:" .. first)
		return
	else
		this:Hide()
		lib.tooltip.currentid = nil
	end
end)

local function _getObjectiveObjectTable()
	local t = table.new();
	t.got  = 0;
	t.string = "";
	t.complete = false;
	return t;
end

function lib:internal_QUEST_LOG_UPDATE()

	-- The quest log is scanned:
	local selectionId = GetQuestLogSelection();
	local numEntries, numQuests = GetNumQuestLogEntries();

	if (not lib.player) then
		lib.player = GetUnitName("player");
	end
	
	if (not lib.db[lib.player]) then
		lib.db[lib.player] = table.new();
	end
	if (not lib.db[lib.player].questcache) then
		lib.db[lib.player].questcache = table.new();
	end
	
	-- collect header information
	local HeaderName, isCollapsed, HeaderStatus, containCollapsedHeader;
	HeaderStatus = table.new();
	for i = 1, numEntries, 1 do
		if select(5, GetQuestLogTitle(i)) then
			-- header
			HeaderName = GetQuestLogTitle(i);
			isCollapsed = select(6, GetQuestLogTitle(i));
			HeaderStatus[HeaderName] = isCollapsed and 1 or 0;
			if isCollapsed then
				containCollapsedHeader = true;
			end
		end
	end

	if (containCollapsedHeader) then
		-- expand all quest headers during scan
		ExpandQuestHeader(0);
		numEntries = GetNumQuestLogEntries();
	end
	
	-- get a list of all current known quest from the lib for this user
	local knownquests = lib:GetQuestLogUniqeIDs(lib.player);
		
	if numEntries > 0 then
		for id = 1, numEntries do
			SelectQuestLogEntry(id);
			local name, level, tag, group, _, _, _, daily, unique_id = GetQuestLogTitle(id);
			if unique_id ~= 0 then
				-- here we have a quest
				if (tContains(knownquests, unique_id)) then
					--find my index for uid 
					local i;
					for k,v in ipairs(knownquests) do
						if (v == unique_id) then
							table.remove(knownquests, k);
							break;
						end
					end
					
				else
					lib:AddQuest(unique_id, lib.player);
				end

				local q = table.new()
				q.L = level
				q.H = level
				q.M = level

				if group > 0 then
					q.group = group
				end

				if tag then
					q.tag = tag
					if  tag == GROUP then
						q.T = "G"
					elseif  tag == ELITE  then
							q.T = "E"
					elseif  tag == ITEM_HEROIC then
							q.T = "H"
					elseif  tag == RAID  then
							q.T = "R"
					elseif  tag == PVP  then
							q.T = "P"
					elseif  tag == LFG_TYPE_DUNGEON then
							q.T = "D"
					end
				end
				q.Y = daily and 1 or nil
				q.S = "S" .. MINOR_VERSION

				q.O = GetNumQuestLeaderBoards(id)
				if q.O and q.O > 0 then
					for o = 1, q.O do
						--_G.print(GetQuestLogLeaderBoard(o, id))
						local otype = select(2,GetQuestLogLeaderBoard(o, id))
						q["o" .. o] = strupper(tostring(otype):sub(1,1))
					end
				end

				-- update lib.db.StaticQuestData
				--lib:internal_UpdateQuestData(unique_id, lib:internal_PackQuestData(q, unique_id))
			end
		end
	end

	if containCollapsedHeader then
		-- restore header
		for i = numEntries, 1 , -1 do
			if select(5, GetQuestLogTitle(i)) then
				-- header
				HeaderName = GetQuestLogTitle(i);
				if HeaderStatus[HeaderName] == 1 then
					-- collapse header
					CollapseQuestHeader(i);
				end
			end
		end
	end
	SelectQuestLogEntry(selectionId);
	
	-- check for quests we have lost from quest log
	if (table.getn(knownquests) > 0) then
		-- we have lost something
		-- now check how we lost the quest
		for _, uid in ipairs(knownquests) do
			print(uid)
			if (lib.loststatus) then
				print("Quest completed: " .. uid);
				lib.callbacks:Fire("Quest_Finished", uid);
			elseif (lib.loststatus == false) then
				print("Quest abandomed: " .. uid);
				lib.callbacks:Fire("Quest_Abandomed", uid);
			else
				assert(false, "Something get wrong by handling lost quests");
			end
			lib:RemoveQuest(uid, lib.player);
		end
		lib.loststatus = nil;
	end
	update_is_running = nil;
end

function lib:internal_DETECT_GROUP(frame)
	--print("internal_DETECT_GROUP")
	
	if frame.grptype == nil then
		-- first run, set up vars
		frame.grptype = false;
		frame.grpmember = 0;
		lib.db.party = table.new();
	end
	
	local grpmember, grouptype;
 	
	if (GetNumRaidMembers() > 0) then
		if(GetRealNumRaidMembers() ~= GetNumRaidMembers()) then
			grouptype = "BATTLEGROUND";
		else
			grouptype = "RAID";
		end
		grpmember = GetNumRaidMembers();
	elseif (GetNumPartyMembers() > 0) then
		--if (GetRealNumPartyMembers() ~= GetNumPartyMembers()) then
			grouptype = "PARTY";
		--else
		--	grouptype = "PARTY";
		--end
		grpmember = GetNumPartyMembers();
	else
		grouptype = false;
		grpmember = 0;
	end
	
	-- debug
	if (grpmember == 0) then
		print("You are not in a party/raid anymore");
	else
		print("You are in a ".. grouptype .. " with " .. (GetNumPartyMembers()) .. " Members");
	end
	
	-- prepare userdata tables
	local member = table.new();
	for k in pairs(lib.db.party) do
		member[k] = true;
	end

	local username, userrealm;
	
	if grpmember > 0 then
		if (grouptype == "RAID" or grouptype == "BATTLEGROUND" ) then
			for i=1, grpmember, 1 do
				username = GetRaidRosterInfo(i);
				if (member[username] == nil) then
					-- user is new
					member[username] = false;
				else
					-- user is already in group, no ation required
					member[username] = nil;
				end
			end
		else
 			for i=1, grpmember, 1 do
				username, userrealm = UnitName("party" .. i);
				if userrealm then
					username = username .. "-" .. userrealm;
				end
				
				if (member[username] == nil) then
					-- user is new
					member[username] = false;
				else
					-- user is already in group, no ation required
					member[username] = nil;
				end
			end
		end
	else
		-- no group, kill vars
		table.deepDel(lib.db.party);
		lib.db.party = table.new();
	end
	
	i = 0;
	for username, userstatus in pairs(member) do
		if (userstatus == true) then
			-- user left group
			-- truncate data
			lib.callbacks:Fire("Party_User_Left", username);
			table.deepDel(lib.db.party[username]);
		else
			-- new in group, init datatables
			print("Add Userdata for " .. username);
			lib.callbacks:Fire("Party_User_Joined", username);
			local userdata = table.new();
			userdata.announcetime = random(200, 800) / 100.0;
			userdata.announcetimeout = 10;
			userdata.quests = table.new();
			userdata.objectives = table.new();
			lib.db.party[username] = userdata;
			i = i + 1;
		end
	end
	
	-- announce
	if i > 1 then
		-- invite in an existing group, say hello to everyone
		lib:SendAddonMessage("Hello", grptype);
	else
		for username in pairs(member) do
			lib:SendAddonMessage("Hello", "WHISPER", username);
		end
	end
end

--
-- OnUpdate Stuff end
--

--
-- Comm Stuff
--

local commhandlers = {
	--version
	--  implicit sync request; 'v' is never broadcast to the party, only to unknown party members
	--  upon receiving 'v', begin a sync to the sender
	v = function(msg, distribution, sender)
		local version = tonumber(msg)
		if version <= last_incompatible_version then
			lib.callbacks:Fire("Incompatible_Version", sender, version)
		end
		if version > MINOR_VERSION then
			lib.callbacks:Fire("New_Version", sender, version)
		end
		lib.worth_sending = true

		lib.party[sender] = version
		--And sync:
		for _, uid in pairs(lib.quest_ids) do
			lib:SendAddonMessage("q"..uid, "WHISPER", sender, "BULK")
			if lib.quest_objectives[uid] then
				for desc, o in pairs(lib.quest_objectives[uid]) do
					local isrep = o[3] == "reputation"
					local got = isrep and lib:GetReactionLevel(o[1]) or o[1] or '1'
					local need = isrep and lib:GetReactionLevel(o[2]) or o[2] or '1'
					lib:SendAddonMessage("o"..uid.."~"..desc.."~"..(isrep and "r" or "")..got.."/"..need, "WHISPER", sender, "BULK")
				end
			end
			lib:SendAddonMessage("r"..uid, "WHISPER", sender, "BULK")
		end
		lib:SendAddonMessage("d", "WHISPER", sender, "BULK")
	end,
	--quest gained
	q = function(msg, distribution, sender)
		local uid = tonumber(msg)
		if not lib.party_quests[sender] then return end
		local name = lib.db.UID_to_name[uid] -- Forces the tooltip request.
		lib.party_quests[sender][uid] = lib.party_quests[sender][uid] or new()
	end,
	--objective update
	o = function(msg, distribution, sender)
		local uid, objective, isrep, got, need = msg:match("^(.-)~(.-)~(r?)(%d-)/(%d-)$")
		uid = tonumber(uid)
		if not (lib.party_quests[sender] and lib.party_quests[sender][uid]) then return end
		local o = lib.party_quests[sender][uid][objective]
		isrep = isrep == 'r'
		got = isrep and lib:GetReactionName(tonumber(got)) or tonumber(got)
		need = isrep and lib:GetReactionName(tonumber(need)) or tonumber(need)
		if o then
			o[1] = got
			o[2] = need
			o[3] = isrep
			if not lib.party_quests[sender].waiting then
				lib.callbacks:Fire("Party_Objective_Update", sender, uid, lib.db.UID_to_name[uid], objective, got, need, isrep)
				lib.callbacks:Fire("Party_Update", sender, uid)
			end
		else
			lib.party_quests[sender][uid][objective] = new(got, need, isrep)
		end
	end,
	--sending a quest is complete
	r = function(msg, distribution, sender)
		local uid = tonumber(msg)
		if not (lib.party_quests[sender] and lib.party_quests[sender][uid]) then return end
		if not lib.party_quests[sender].waiting then
			lib.callbacks:Fire("Party_Quest_Gained", sender, uid, lib.db.UID_to_name[uid])
			lib.callbacks:Fire("Party_Update", sender, uid)
		end
	end,
	--quest finished
	c = function(msg, distribution, sender)
		local uid = tonumber(msg)
		lib.callbacks:Fire("Party_Quest_Complete", sender, uid, lib.db.UID_to_name[uid])
		lib.callbacks:Fire("Party_Update", sender, uid)
	end,
	--quest failed
	f = function(msg, distribution, sender)
		local uid = tonumber(msg)
		lib.callbacks:Fire("Party_Quest_Failed", sender, uid, lib.db.UID_to_name[uid])
		lib.callbacks:Fire("Party_Update", sender, uid)
	end,
	--quest lost
	l = function(msg, distribution, sender)
		local uid = tonumber(msg)
		if not (lib.party_quests[sender] and lib.party_quests[sender][uid]) then return end
		lib.callbacks:Fire("Party_Quest_Lost", sender, uid, lib.db.UID_to_name[uid])
		lib.callbacks:Fire("Party_Update", sender, uid)
		lib.party_quests[sender][uid] = deepDel(lib.party_quests[sender][uid])
	end,
	--sync done
	d = function(msg, distribution, sender)
		if not lib.party_quests[sender] then return end
		lib.party_quests[sender].waiting = nil
		lib.callbacks:Fire("Sync_Finished", sender)
		lib.callbacks:Fire("Party_Update", sender)
	end,
}

-- Comms:
-- Someone joins party -> version info exchanged -> quest data shared

local last_incompatible_version = 0

--lib.party = lib.party or table.new()
--lib.party_quests = lib.party_quests or table.new()

local ChatThrottleLib = ChatThrottleLib
function lib:SendAddonMessage(contents, distribution, target, priority)
	if ChatThrottleLib and not lib.disable_comms then
		ChatThrottleLib:SendAddonMessage(priority or "NORMAL", "lQI", contents, distribution or "PARTY", target);
	end
end

--
-- Comm Stuff end
--

--
-- Internal API
--

function lib:internal_PackQuestData(datatable, qid)
	if (not datatable or not qid or type(datatable) ~= "table" or type(qid) ~= "number") then
		assert(false, "Usage: LibQuestInfo:PackQuestData([table] datatable, [number] uid)");
	end
	

	local queststring = "Q"..tostring(qid) .. ";"

	for _, k in ipairs(datakeys) do
		--print(k .. " " .. v)
		if datatable[k] then
			if k == "O" and datatable[k] > 0 then
				queststring = queststring .. "O." .. tostring(datatable[k]) .. ";"
				for i=1, datatable[k] do
					--print(i .. " " .. tostring(i) .. " " .. tostring(datatable["o"..i]))
					queststring = queststring .. "o" .. tostring(i) .. "."..  tostring(datatable["o"..i]) .. ";"
				end
			elseif k =="T" then
				if datatable[k] == "G" and datatable.group and datatable.group > 0 then
					queststring = queststring .. k .."." .. tostring(datatable[k]) .. "." .. tostring(datatable.group) .. ";"
				else
					queststring = queststring .. k .."." .. tostring(datatable[k]) .. ";"
				end
			elseif k == "L" or (k == "M" and datatable.L ~= datatable.M) or (k == "H" and datatable.L ~= datatable.H) or k == "Y" or k =="W" or k == "S" then
				queststring = queststring .. k .."." .. tostring(datatable[k]) .. ";"
			end
		end
	end
	return queststring .. "|"
end

function lib:internal_UnpackQuestData(datastring)
	if (not datastring or type(datastring) ~= "string") then
		assert(false, "Usage: LibQuestInfo:UnpackQuestData([string] datastring)");
	end
	
	-- Split String in all avalible elements
	local tmptable = table.new({strsplit(";", datastring)})

	local datatable = table.new()
	for _, v in ipairs(tmptable) do
		local key, data, group = strsplit(".", v)
		if (group) then
			datatable["group"] = tonumber(group)
		end
		
		-- is it an objective?
		if (strsub(key, 1, 2) == "o" ) then
			-- data comes in @ o1.I,8
			datatable[key] = table.new();
			datatable[key] = tonumber(data);
		else
			datatable[key] = data;
		end
	end
	
	-- calc amount objectives
	for i = 1, true, 1 do
		if (not datatable["o" .. i]) then
			if (i > 1) then
				datatable.O = i -1;
			end
			break;
		end
	end

	return datatable
end

function lib:internal_LoadQuestData(qid)
	if (not qid or type(qid) ~= 'number') then
		assert(false, "Usage: LibQuestInfo:LoadQuestData([number] QuestID)");
	end

	if lib.db.Cache.QuestCache[qid] then
		
		if (not lib.db.StaticQuestData[ceil(qid/100)]) then
			-- we do not have valid Data in our DB for this questrange
			return false;
		end
		
		-- get Datastring from DB
		local questdata = string.match(lib.db.StaticQuestData[ceil(qid/100)], "|Q"..qid..";(.-)|");

		if not questdata then
			-- we have no valid Data in our DB for this Quest
			return false;
		end
		
		-- Unpack Data
		lib.db.Cache.QuestCache[qid] = lib:UnpackQuestData(questdata);
	end

	-- Data loaded sucessfully
	return true;
end

function lib:internal_InitCacheData()

	for catkey, catvalue in pairs() do
		if type(catvalue) == 'number' then
			-- Category Key is a Criteria of an achievement
			table.insert(lib.db.Cache.QuestCategoryCache, GetAchievementCriteriaInfo(catvalue))
		elseif true then
			
		end
	end

end

function lib:internal_UpdateQuestData(qid, newquestdata)
	if not lib.db.StaticQuestData[ceil(qid/100)] then
		lib.db.StaticQuestData[ceil(qid/100)] = '';
	end

	if string.match(lib.db.StaticQuestData[ceil(qid/100)], newquestdata) then
		-- quest not updated -> exit
		return false
	end

	local oldquestdata = string.match(lib.db.StaticQuestData[ceil(qid/100)], "|(Q"..qid..";.-|)")

		-- quest doesn't exist in quest db, insert and exit
	if not oldquestdata then
		lib.db.StaticQuestData[ceil(qid/100)] = lib.db.StaticQuestData[ceil(qid/100)] .. newquestdata
		lib.callbacks:Fire("QuestDataUpdate", qid, newquestdata)
		return true
	end

	local newquest = lib:UnpackQuestData(newquestdata)
	local oldquest = lib:UnpackQuestData(oldquestdata)

	-- calc priority of questdata, bases on version of lib and source
	oldquest.prio = (priotable[string.match(oldquest.S, "(%a)")] or -1) + (string.match(oldquest.S, "(%d)") or 0)
	newquest.prio = (priotable[string.match(newquest.S, "(%a)")] or -1) + (string.match(newquest.S, "(%d)") or 0)

	local quest, priolevel
	if oldquest.prio > newquest.prio then
		priolevel = oldquest.prio - newquest.prio
		if priolevel == 1 then
			-- data is not to old, do a key based comparsion
			quest = CopyTable(oldquest)
		else
			return false
		end

	else
		priolevel = newquest.prio - oldquest.prio
		if priolevel == 1 or priolevel == 0 then
			-- data be is not to old, do a key based comparsion
			quest = CopyTable(newquest)
		else
			-- new data is newer, insert and exit
			lib.db.StaticQuestData[ceil(qid/100)] = lib.db.StaticQuestData[ceil(qid/100)] .. newquestdata
			lib.callbacks:Fire("QuestDataUpdate", qid, newquestdata)
			return true
		end
	end


	local changed

	print("Do extended Check for " .. qid)
	for _, key in ipairs({"O","L","M","H","Y","W","T","group"}) do
		if newquest[key] ~= oldquest[key] then
			if not changed then
				changed = table.new()
			end
			if key == "M" or key == "H" then
				changed.L = true
			else
				changed[key] = true
			end
		end
	end

	-- data not changed, but order is different
	if not changed then
		print("Data for " .. qid .. " not changed (" .. tostring(changed) .. ")")
		return false
	end




	--print(prolevel)
	for k in pairs(changed) do
		-- level
		if k == "L" then
			if not newquest.L or not oldquest.L then
				-- one quest has no leveldata, use leveldata from existning quest
				quest.L = newquest.L or oldquest.L
			else
				-- both quest have a level, set min/max questlevel
				-- this have nothing to do with gained XP Range
				quest.L = -1
				quest.M = min(oldquest.M, newquest.M)
				quest.H = max(oldquest.H, newquest.H)
			end
		elseif priolevel <= 1 then
			quest[k] = oldquest[k] or newquest[k] or nil
		end
	end


	lib.callbacks:Fire("QuestDataUpdate", qid, lib:PackQuestData(quest))

	return true
end

-- Public API:

function lib:GetQuestLogTitleByID(qid, silent)
	if (not qid or type(qid) ~= 'number') then
		if not silent then
			assert(false, "Usage: LibQuestInfo:GetQuestLogTitleByID([number] uid)");
		end
		return
	end

	if not lib.db.Cache.QuestCache[qid] or not lib.db.Cache.QuestCache[qid].title then
		lib:internal_LoadQuestData(qid);
		--print("Quest not loaded " .. qid)
		return nil, 0
	end

	-- questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, qid, isLoaded
	return lib:GetQuestName(qid), questcache[qid].L or 0, questcache[qid].T and TagTypes[questcache[qid].T] or nil, questcache[qid].group or 0, nil, nil, nil, questcache[qid].Y, qid
end

-- Returns the ID of Quests Category
function lib:GetQuestCategoryID(QuestID)
	if (not QuestID or type(QuestID) ~= "number") then
		assert(false, "Usage: LibQuestInfo:GetQuestCategoryID([number] QuestID)");
	end
	
	return lib.db.Cache.QuestCache[QuestID] and lib.db.Cache.QuestCache[QuestID]['ZoneID'] and lib.db.Cache.QuestCache[QuestID]['ZoneID'] or (ceil(QuestID/100));
end

-- Returns the Name of Quests Category
function lib:GetQuestCategoryName(QuestID)
	if (not QuestID or type(QuestID) ~= "number") then
		assert(false, "Usage: LibQuestInfo:GetQuestCategoryName([number] QuestID)");
	end
	local id = lib:GetQuestCategoryID(QuestID);
	return lib.db.Cache.CategoryCache[id] or "UNKNOWN";
end

function lib:GetQuestName(qid, tmp)
	if (not qid or type(qid) ~= "number") then
		assert(false, "Usage: LibQuestInfo:GetQuestName([number] uid)");
	end

	if not questcache[qid] or not questcache[qid].title then
		if not tmp then
			if not loadqueue[qid] then
				loadqueue[qid] = true
			end
			lib.frame:Show()
			return nil
		end
		return nil
	else
		return questcache[qid].title
	end
end

function lib:GetHighestQuestNr()
	local HiQid = 0;
	local tables = {"StaticQuestData", "UpdatedQuestData"}
	
	for k in pairs(lib.db.StaticQuestData) do
		HiQid = max(HiQid, k);
	end
	
	for k in pairs(lib.db.UpdatedQuestData) do
		HiQid = max(HiQid, k);
	end
	
	local qid = 0;
	if lib.db.StaticQuestData[HiQid] then
		local qdata = { strsplit("|", lib.db.StaticQuestData[HiQid]) };
		for key, value in next, qdata do
			qid = max(qid, tonumber(string.match( value, "^Q(%d*);")));
		end
	end
	
	if lib.db.UpdatedQuestData[HiQid] then
		local qdata = { strsplit("|", lib.db.UpdatedQuestData[HiQid]) };
		for key, value in next, qdata do
			qid = max(qid, tonumber(string.match( value, "^Q(%d*);")));
		end
	end

	return qid;
end

-- Get a numeric equivalent to a reaction level, from Hated=1 to Exalted=8.
do
	local reactions = {
		[FACTION_STANDING_LABEL1] = 1,
		[FACTION_STANDING_LABEL1_FEMALE] = 1,
		[FACTION_STANDING_LABEL2] = 2,
		[FACTION_STANDING_LABEL2_FEMALE] = 2,
		[FACTION_STANDING_LABEL3] = 3,
		[FACTION_STANDING_LABEL3_FEMALE] = 3,
		[FACTION_STANDING_LABEL4] = 4,
		[FACTION_STANDING_LABEL4_FEMALE] = 4,
		[FACTION_STANDING_LABEL5] = 5,
		[FACTION_STANDING_LABEL5_FEMALE] = 5,
		[FACTION_STANDING_LABEL6] = 6,
		[FACTION_STANDING_LABEL6_FEMALE] = 6,
		[FACTION_STANDING_LABEL7] = 7,
		[FACTION_STANDING_LABEL7_FEMALE] = 7,
		[FACTION_STANDING_LABEL8] = 8,
		[FACTION_STANDING_LABEL8_FEMALE] = 8,
	}
	--------------------------------------------
	-- Arguments:
	-- string - faction standing text
	--
	-- Notes:
	-- Returns a numeric-equivalent for faction standings.
	--
	-- Returns:
	-- * number - number from 1 to 8 or -1 if invalid
	--------------------------------------------
	function lib:GetReactionLevel(leveltext)
		return leveltext and reactions[leveltext] or -1
	end

	function lib:GetReactionName(level)
		for react, index in pairs(reactions) do
			if index == level then
				return react
			end
		end
	end
end


-- Stuff for party 


local ignoreupdate 

function lib:AddQuest(questdata, username)
	-- check parameter vars
	if ( not username or not questdata or type(username) ~= "string" or type(questdata) ~= "number" and type(questdata) ~= "table" ) then 
		assert(false, "Usage: LibQuestInfo:AddQuest([number|table] unique questid, [string] username)");
	end
	
	if lib.db[username] == nil then
		-- benutzertabelle nicht vorhanden
		lib.db[username] = table.new();
		lib.db[username].questcache = table.new();
		lib.db[username].progresscache = table.new();
		lib.db[username].headerstatus = table.new();
	end
	
	if type(questdata) == 'table' then
		ignoreupdate = true;
		for k, v in pairs(questdata) do
			if v == true then
				-- data from questquerycomplete, use key
				lib:AddQuest(k, username);
			else
				-- use value
				lib:AddQuest(v, username);
			end
		end
		ignoreupdate = nil;
	elseif (not lib:UserOnQuest(questdata, username)) then
		table.insert(lib.db[username].questcache, questdata);
		-- fire callback
		lib.callbacks:Fire("Quest_Accepted", questdata);
	end
	
	if not ignoreupdate then 
		lib:internal_SortQuestTable(username);
	end
end

function lib:RemoveQuest(questdata, username)
	-- check parameter vars
	if ( not username or not questdata or type(username) ~= "string" or type(questdata) ~= "number" ) then 
		assert(false, "Usage: LibQuestInfo:RemoveQuest([number] unique questid, [string] username)");
	end
	
	
	for k,v in ipairs(lib.db[username].questcache) do
		
		-- quest found?
		if v == questdata then
			if (type(v) ~= "string") then
				-- string is header, not a quest
				if ( type(lib.db[username][k-1]) == "string" and
						( -- tricky, kill header only if
							lib.db[username][k+1]						-- next key exist 
							and type(lib.db[username][k+1]) == "string"	-- and is a header
							or not lib.db[username][k+1]				-- or does not exist
						)
					) then
					-- header - quest - header (or empty)
					-- remove header and quest from list
					k = k - 1;
					v = lib.db[username][k] * -1;
					
					-- zonenkey lschen (achtung, wert verschiebt sich um eine stelle nach oben)
					table.remove(lib.db[username].questcache, k);
					-- wert lschen
					table.remove(lib.db[username].questcache, k);
					-- berschriftstatus lschen
					lib:internal_RemoveHeaderStatus(username, v) --todo
				else
					-- delete quest only
					table.remove(lib.db[username].questcache, k);
				end
			end
		end
	end
end

-- local sort function
local function _sort(a,b)
	local title, level, zone
	local queststringa, queststringb = nil, nil;
	
	-- check quest 1 is cached
	if (not lib.db.SortCache[a]) then
		-- quest not loaded or cached, try load and cache it
		title, level = lib:GetQuestLogTitleByID(a);
		zone = lib:GetQuestCategoryName(a);
		if (zone == "UNKNOWN" or not title or level == 0) then
			-- quest data is not avalible, create a dummy
			queststringa = zone .. "|" .. level .. "|" .. a
		else
			-- quest data is avalible, save it to cache
			table.insert(lib.db.SortCache, a, (zone .. "|" .. level .. "|" .. title));
		end
	else
		queststringa = lib.db.SortCache[a]
	end
	
	-- check quest 2 is cached
	if (not lib.db.SortCache[b]) then
		-- quest not loaded or cached, try load and cache it
		title, level = lib:GetQuestLogTitleByID(b);
		zone = lib:GetQuestCategoryName(b);
		if (zone == "UNKNOWN" or not title or level == 0) then
			-- quest data is not avalible, create a dummy
			queststringb = zone .. "|" .. level .. "|" .. b
		else
			-- quest data is avalible, save it to cache
			table.insert(lib.db.SortCache, b, (zone .. "|" .. level .. "|" .. title));
		end
	else
		queststringb = lib.db.SortCache[b]
	end
	
	-- preparation stuff is done
	-- compare!
	return (queststringa  > queststringb)
end

function lib:internal_SortQuestTable(username)
	if not username or type(username) ~= "string" then
		assert(false, "Usage: LibQuestInfo:internal_SortQuestTable([string] username)");
	end

	local questlist = lib.db[username].questcache;
	--print(table.concat(questlist, " "));
	-- remove header
	for i = table.getn(questlist), 1, -1 do
		if (type(questlist[i]) == "string") then
			table.remove(questlist, i)
		end
	end
	
	-- sort table
	--print(table.concat(questlist, " "));
	table.sort(questlist, _sort);
	--print(table.concat(questlist, " "));
	
	-- insert header
	local lastzone = nil
	for i = table.getn(questlist), 1, -1 do
		-- ausnahmen: erster durchlauf
		if lastzone == nil then
			lastzone = tostring(lib:GetQuestCategoryID(questlist[i]));
		end
		
		if (lastzone ~= tostring(lib:GetQuestCategoryID(questlist[i]))) then
			-- Zone vor der Letzten Quest der Zone einfgen
			table.insert(questlist, i + 1 ,lastzone)
			lastzone = tostring(lib:GetQuestCategoryID(questlist[i]));
		end
		
		-- ausnahme: letzte quest
		if i == 1 then
			table.insert(questlist, i, tostring(lib:GetQuestCategoryID(questlist[i])));
		end
	end
	--print(table.concat(questlist, " "));
end

-- Public Party API

function lib:CollapseQuestHeader(questindex, username)
	if (not username or not questindex or type(username) ~= "string" or type(questindex) ~= "number" ) then
		assert(false, "Usage: LibQuestInfo:GetNumQuestLogEntries([number] questindex[string] username)");
	end
	
	if username == nil then 
		-- keine Username gesetzt, nehme aktiven User
		username = UnitName("player");
	end
		
	assert(false, "to implement")
end

function lib:ExpandQuestHeader(questindex, username)
	if username == nil then 
		-- keine Username gesetzt, nehme aktiven User
		username = UnitName("player");
	end
	assert(false, "to implement")
end

function lib:GetNumQuestLogEntries(username)
	if (not username or type(username) ~= "string") then
		assert(false, "Usage: LibQuestInfo:GetNumQuestLogEntries([string] username)");
	end

	local numQuest, numTotal = 0, 0;
	for k, uid in pairs(lib.db[username].questcache) do
		numTotal = numTotal + 1;
		if (type(uid) == "number") then
			numQuests = numQuests + 1;
		end
	end
	-- ToDo: collapsed Status
	return numTotal, numQuests;
end

function lib:GetQuestLogQuestText(uid)
	if (not uid or type(uid) ~= "number") then
		assert(false, "Usage: LibQuestInfo:GetQuestLogQuestText([number] uid)");
	end

	return lib.db.Cache.QuestCache[uid] and lib.db.Cache.QuestCache[uid].Text or "";
end

function lib:GetQuestLogTitle(questindex, username)
	if (not questindex or not username or type(questindex) ~= "number" or type(username) ~= "string") then 
		assert(false, "Usage: LibQuestInfo:GetQuestLogUniqeIDs([number] questindex, [string] username)");
	end
	
	-- check questindx
	if questindex < 1 then
		assert(false, "LibQuestInfo:GetQuestLogUniqeIDs: questindex must be a number greather then 0");
		return;
	end

	
	-- search for index in table
	if not lib.db[username] or not lib.db[username][questindex] then
		return nil;
	end
	
	local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID
	
	questID = lib.db[username][questindex];
	
	if (type(questID) == "string") then
		-- it is a category header
		isHeader = 1;
	else
		
	end
	assert(false, "to implement")
end

function lib:GetQuestLogUniqeIDs(username)
	if (not username or type(username) ~= "string") then
		assert(false, "Usage: LibQuestInfo:GetQuestLogUniqeIDs([string] username)");
	end

	local t = table.new();
	
	-- remove headers
	for k, uid in pairs(lib.db[username].questcache) do
		if type(uid) == "number" and uid > 0 then
			-- questid
			table.insert(t,uid);
		end
	end
	return t;
end

function lib:UserOnQuest(uid, username)
	if (not username or  not uid or type(username) ~= "string" or type(uid) ~= "number") then 
		-- keine Username gesetzt
		assert(false, "Usage: LibQuestInfo:UserOnQuest([number] unique questid, [string] username)");
	end

	return tContains(lib.db[username].questcache, uid);
end

-- Public Questinfo API

function lib:GetNumQuestLeaderBoards(uid)
	if (not uid or type(uid) ~= "number") then
		assert(false, "Usage: LibQuestInfo:GetNumQuestLeaderBoards([number] uid)");
	end
	
	local i, k = 0;
	for k = 1, true, 1 do
		if (not lib.db.Cache.QuestCache[uid]["o" .. k]) then
			break;
		end
		i = i + 1;
	end
	
	return i;
end

function lib:GetQuestLogLeaderBoard(uid, oindex, username)
	if (not uid or not oindex or type(uid) ~= "number" or type(oindex) ~= "number" ) then
		assert(false, "Usage: LibQuestInfo:GetQuestLogLeaderBoard([number] uid, [number] oindex[, [string] username])");
	end
	assert(false, "to implement")
end

-- debug
lib:RegisterCallback("Quest_Added", function(event, title, uid, objectives)

end)


